package GT::Maildir::FolderTree::Dir;

use vars qw/$error $ERRORS $ERROR_MESSAGE/;
use base 'GT::Base';
use strict;
use warnings;

use GT::Maildir qw/:status/;

use Carp qw/croak/;

$ERRORS = {
};
$ERROR_MESSAGE = "GT::Maildir::FolderTree";

sub new {
    my $class = shift;
    my $self = $class->SUPER::new;
    croak "Invalid arguments to $class->new: ", join(", ", @_) if @_ & 1;
    my %opts = @_;
    $opts{ lc $_ } = delete $opts{$_} for keys %opts;

    my $folders = delete $opts{folders};
    croak "No folder object specified to $class->new" unless $folders;
    my $directory = delete $opts{directory};
    croak "No directory to list specified to $class->new" unless defined $directory and length $directory;
    my $list_cur = delete $opts{listcurrentmessages};
    my $list_new = delete $opts{listnewmessages};
    my $list_folders = delete $opts{listfolders};
    my $search_index = delete $opts{searchindex};
    my $search_sort = delete $opts{searchsort};
    my $offset = delete $opts{offset};
    my $limit = delete $opts{limit};
    if (!$list_cur and !$list_new and !$list_folders) {
        croak "Nothing specified to list to $class->new";
    }
    if (!$list_cur and !$list_new) {
        $search_index = $search_sort = undef;
    }
    $self->{_debug} = delete $opts{debug};
    croak "Unknown arguments to $class->new: ", join(", ", map { $_ . " => " . $opts{$_} } keys %opts) if keys %opts;
    $self->{folders} = $folders;
    $self->{directory} = $directory;
    $self->{list_cur} = $list_cur;
    $self->{list_new} = $list_new;
    $self->{list_folders} = $list_folders;
    $self->{search_index} = $search_index;
    $self->{search_sort} = $search_sort;
    $self->{offset} = $offset;
    $self->{limit} = $limit;
    $self->{pos} = 0;
    return $self;
}

sub readdir {
    my $self = shift;
    croak "Unknown arguments to $self->readdir: ", join(", ", @_) if @_;
    if (defined $self->{limit}) {
        if ($self->{pos} >= $self->{offset} + $self->{limit}) {
            _close_all($self);
            return undef;
        }
    }
    if ($self->{list_folders} and !_is_folders_done($self)) {
        if (!_is_folders_open($self)) {
            _open_folders($self) or return undef;
            $self->{pos} = _seek_folders($self, $self->{offset}) if $self->{offset};
        }
        my $next_folder = _next_folder($self);
        if (defined $next_folder) {
            $self->{pos}++;
            return $next_folder;
        }
        _close_folders($self);
    }
    if ($self->{search_index}) {
        if (!_is_search_done($self)) {
            if (!_is_search_open($self)) {
                _open_search($self) or return undef;
                if ($self->{offset}) {
                    my $seek = $self->{offset} - $self->{pos};
                    if ($seek > 0) {
                        $self->{pos} = _seek_search($self, $seek);
                    }
                }
            }
            my $data = _next_search($self);
            if (defined $data) {
                $self->{pos}++;
                return $data;
            }
            _close_search($self);
        }
    }
    else {
        if ($self->{list_new} and !_is_dir_done($self, "new")) {
            if (!_is_dir_open($self, "new")) {
                _open_dir($self, "new") or return undef;
                if ($self->{offset}) {
                    my $seek = $self->{offset} - $self->{pos};
                    if ($seek > 0) {
                        $self->{pos} = _seek_dir($self, "new", $seek);
                    }
                }
            }
            my $path = _next_dir($self, "new");
            if (defined $path) {
                $self->{pos}++;
                return $path;
            }
            _close_dir($self, "new");
        }
        if ($self->{list_cur} and !_is_dir_done($self, "cur")) {
            if (!_is_dir_open($self, "cur")) {
                _open_dir($self, "cur") or return undef;
                if ($self->{offset}) {
                    my $seek = $self->{offset} - $self->{pos};
                    if ($seek > 0) {
                        $self->{pos} = _seek_dir($self, "cur", $seek);
                    }
                }
            }
            my $path = _next_dir($self, "cur");
            if (defined $path) {
                $self->{pos}++;
                return $path;
            }
            _close_dir($self, "cur");
        }
    }
    return undef;
}

sub rewinddir {
    my $self = shift;
    croak "Unknown arguments to $self->rewinddir: ", join(", ", @_) if @_;
    _close_all($self);
    return 1;
}

sub telldir {
    my $self = shift;
    croak "Unknown arguments to $self->telldir: ", join(", ", @_) if @_;
    return $self->{pos};
}

sub _seek_dir {
    my ($self, $which, $pos) = @_;
    my $dh = $self->{"${which}_dir"};
    seekdir $dh, 0;
    my $i = 0;
    while ($i < $pos) {
        _dir_next($self, $which) or last;
        $i++;
    }
    $self->{"${which}_pos"} = $i;
    return $i;
}

sub _seek_search {
    my ($self, $pos) = @_;
    my $last_idx = $#{$self->{search_results}};
    $self->{search_pos} = $pos > $last_idx ? $last_idx : $pos;
    return $self->{search_pos};
}

sub _seek_folders {
    my ($self, $pos) = @_;
    my $last_idx = $#{$self->{folders_list}};
    $self->{folders_pos} = $pos > $last_idx ? $last_idx : $pos;
    return $self->{folders_pos};
}

sub _close_dir {
    my ($self, $which) = @_;
    if ($self->{"${which}_dir"}) {
        closedir $self->{"${which}_dir"};
        $self->{"${which}_dir"} = undef;
    }
    $self->{"${which}_pos"} = 0;
    $self->{"list_${which}_done"} = 1;
}

sub _close_search {
    my ($self) = @_;
    $self->{search_results} = undef;
    $self->{search_pos} = 0;
    $self->{list_search_done} = 1;
}

sub _close_folders {
    my ($self) = @_;
    $self->{folder_list} = undef;
    $self->{folders_pos} = 0;
    $self->{list_folders_done} = 1;
}

sub _close_all {
    my $self = shift;
    _close_dir($self, "new");
    _close_dir($self, "cur");
    _close_search($self);
    _close_folders($self);
    $self->{pos} = $self->{offset};
}

sub _open_dir {
    my ($self, $which) = @_;
    my $folder = $self->{folders}->get_folder($self->{directory});
    my $meth = "get_${which}_path";
    my $path = $folder->$meth();
    $self->{"${which}_pos"} = 0;
    my $dh = do {
        my $class = ref $self;
        no strict 'refs';
        *{"$class\:\:DIR_\U${which}\E"};
    };
    opendir $dh, $path or return undef;
    $self->{"${which}_dir"} = $dh;
}

sub _open_search {
    my ($self) = @_;
    require GT::Maildir::Search::Query;
    require GT::Maildir::Search::Term;
    $self->{search_pos} = 0;
    if (!$self->{list_cur} and !$self->{list_new}) {
        $self->{search_results} = [];
        return 1;
    }
    my @terms;
    push @terms, GT::Maildir::Search::Term->new(
        Field => "folder",
        Query => qq("$self->{directory}")
    );
    if (!$self->{list_cur} or !$self->{list_new}) {
        if ($self->{list_cur}) {
            push @terms, GT::Maildir::Search::Term->new(
                Field => "status",
                Query => ST_CUR
            );
        }
        elsif ($self->{list_new}) {
            push @terms, GT::Maildir::Search::Term->new(
                Field => "status",
                Query => ST_NEW
            );
        }
    }
    my %opts;
    if ($self->{search_sort}) {
        $opts{Sort} = $self->{search_sort};
    }
    $opts{Query} = GT::Maildir::Search::Query->new(
        Bool => GT::Maildir::Search::Query::GQUERY_AND(),
        Terms => \@terms
    );
    $self->{search_results} = $self->{search_index}->search(%opts) or return $self->error($GT::Maildir::Searcher::error);
    return 1;
}

sub _open_folders {
    my ($self) = @_;
    $self->{folders_pos} = 0;
    $self->{folders_list} = [];
    my $folders = $self->{folders}->get_list || [];
    my $match = $self->{directory};
    $match .= '.' unless $match eq '.';
    my %uniq;
    while (@$folders) {
        my $next_folder = shift @$folders;
        if ($next_folder =~ /^\Q$match\E([^.]*)/) {
            next if $uniq{$1}++;
            push @{$self->{folders_list}}, ("$1" || '.');
        }
    }
    return 1;
}

sub _is_dir_open {
    my ($self, $which) = @_;
    return defined $self->{"${which}_dir"};
}

sub _is_folders_open {
    my ($self) = @_;
    return defined $self->{folders_list};
}

sub _is_search_open {
    my ($self) = @_;
    return defined $self->{search_results};
}

sub _is_dir_done {
    my ($self, $which) = @_;
    return $self->{"list_${which}_done"};
}

sub _is_search_done {
    my ($self) = @_;
    return $self->{list_search_done};
}

sub _is_folders_done {
    my ($self) = @_;
    return $self->{list_folders_done};
}

sub _next_dir {
    my ($self, $which) = @_;
    my $message = CORE::readdir $self->{"${which}_dir"};
    while (defined $message and ($message eq "." or $message eq "..")) {
        $message = CORE::readdir $self->{"${which}_dir"};
    }
    if (defined $message) {
        return "$which/$message";
    }
    return undef;
}

sub _next_search {
    my ($self) = @_;
    if ($self->{search_pos} > $#{$self->{search_results}}) {
        return undef;
    }
    return $self->{search_results}[$self->{search_pos}++];
}

sub _next_folder {
    my ($self) = @_;
    if ($self->{folders_pos} > $#{$self->{folders_list}}) {
        return undef;
    }
    return $self->{folders_list}[$self->{folders_pos}++];
}

1;

