package GT::Maildir::Lucene;

use strict;
use warnings;
use vars qw/%Lucene_Table_Defs @EXPORT_OK %EXPORT_TAGS $Default_Field %Stores $error/;

use Carp qw/croak/;
use Exporter();

*import = \&Exporter::import;

sub SB_DOCSCORE () { 0 }
sub SB_DOC      () { 1 }
sub SB_AUTO     () { 2 }
sub SB_STRING   () { 3 }
sub SB_INT      () { 4 }
sub SB_FLOAT    () { 5 }

$Default_Field = "subject";

# Table def:
%Lucene_Table_Defs = (
    date => {
        storage => "Keyword", 
        sort_type => SB_INT
    },
    date_recieved => {
        storage => "Keyword", 
        sort_type => SB_INT
    },
    subject => {
        storage => "Text", 
        sort_type => SB_STRING,
        sort_field => "subject_sort"
    },
    subject_sort => {
        storage => "Keyword", 
        sort_type => SB_STRING
    },
    has_attachment => {
        storage => "Keyword", 
        sort_type => SB_INT
    },
    has_html => {
        storage => "Keyword", 
        sort_type => SB_INT
    },
    size => {
        storage => "Keyword", 
        sort_type => SB_INT
    },
    content_type => {
        storage => "Keyword", 
        sort_type => SB_STRING
    },
    spam_status => {
        storage => "Keyword", 
        sort_type => SB_INT
    },
    spam_score => {
        storage => "Keyword",
        sort_type => SB_INT
    },
    spam_score_required => {
        storage => "Keyword",
        sort_type => SB_INT
    },
    spam_learn => {
        storage => "Keyword",
        sort_type => SB_INT
    },
    spam_report => {
        storage => "Keyword", 
        sort_type => SB_STRING
    },
    spam_version => {
        storage => "Keyword", 
        sort_type => SB_STRING
    },
    dangerous => {
        storage => "UnIndexed", 
        sort_type => SB_INT
    },
    body => {
        storage => "Text",
        sort_type => SB_STRING
    },
    body_preview => {
        storage => "UnIndexed",
        sort_type => SB_STRING
    },
    header => {
        storage => "Text",
        sort_type => SB_STRING
    },
    message_id => {
        storage => "Keyword",
        sort_type => SB_STRING
    },
# With these three we can get the path on disk
# Also, these tree make for a unique identifier
    folder => {
        storage => "Keyword", 
        sort_type => SB_STRING
    },
    status => { # new/cur
        storage => "Keyword",
        sort_type => SB_INT
    },
    filename => {
        storage => "UnIndexed", 
        sort_type => SB_STRING
    },
    # A combination of the above three, needed because a delete query can only
    # be on one field ATM...
    path => {
        storage => "Keyword",
        sort_type => SB_STRING
    },
    # Flags, these are stored in the path by courier imap
    flag_answered => {
        storage => "Keyword",
        sort_type => SB_INT
    },
    flag_seen => {
        storage => "Keyword",
        sort_type => SB_INT
    },
    flag_draft => {
        storage => "Keyword",
        sort_type => SB_INT
    },
    flag_deleted => {
        storage => "Keyword",
        sort_type => SB_INT
    },
    data => {
        storage => "Keyword",
        sort_type => SB_STRING
    }
);

for my $field (qw(to from cc bcc reply_to)) {
    for ('', qw(_name _email)) {
        $Lucene_Table_Defs{"${field}${_}"} = {
            storage => "Text",
            sort_type => SB_STRING,
            sort_field => "${field}${_}_sort"
        };
        $Lucene_Table_Defs{"${field}${_}_sort"} = {
            storage => "Keyword",
            sort_type => SB_STRING,
        };
    }
}
@EXPORT_OK = qw(
    SB_DOCSCORE
    SB_DOC
    SB_AUTO
    SB_STRING
    SB_INT
    SB_FLOAT

    %Lucene_Table_Defs
);

%EXPORT_TAGS = (
    all => [@EXPORT_OK],
    sort_type => [qw/SB_DOCSCORE SB_DOC SB_AUTO SB_STRING SB_INT SB_FLOAT/]
);

sub store {
    my $class = shift;
    my $path = shift;
    croak "No path specified to $class\::store()" unless defined $path and length $path;
    croak "Unknown arguments to $class\::store(): ", join(", ", @_) if @_;
    unless (exists $Stores{$path}) {
        my $create = !-e $path;
        my $store = eval {
            local $SIG{__DIE__};
            Lucene::Store::FSDirectory->getDirectory($path, $create);
        };
        if ($@) {
            $GT::Maildir::Lucene::error = "$@";
            return undef;
        }
        $Stores{$path} = $store;
    }
    return $Stores{$path};
}

# Class Method
# Returns a list of the fields in the database. Excludes sort fields
sub get_fields_list {
    my $class = shift;
    croak "Unknown arguments to $class->get_fields_list: ", @_ if @_;
    return sort grep !/_sort$/, keys %Lucene_Table_Defs;
}

sub get_default_field {
    my $class = shift;
    croak "Unknown arguments to $class->get_default_field: ", @_ if @_;
    return $Default_Field;
}

sub get_table_defs {
    my $class = shift;
    croak "Unknown arguments to $class->get_table_definition: ", @_ if @_;
    return {%Lucene_Table_Defs};
}

1;

