package GT::Maildir::Quota;

use strict;
use vars qw/$error $ERRORS $Trash_Quota $Trash_Folder/;
use base 'GT::Base';

use GT::Maildir qw/:stat/;
use GT::TempFile;

use Fcntl qw/:mode/;
use Carp qw/croak/;

sub MAILDIRQUOTA_RECALCULATE_DELAY() { 16 * 60 }

$error = undef;
$Trash_Quota = 0;
$Trash_Folder = "Trash";
$ERRORS = {
    RENAME => "Could not rename %s to %s: %s"
};

sub new {
    my $self = shift;
    croak "Use $self->open(), not $self->new()";
}


sub open {
    my $class = shift;
    my $dir = shift;
    my $quotastr = shift;
    croak "Unknown arguments to $class->open" if @_;
    croak "No directory specified to $class->open" unless defined $dir and length $dir;

    my $self = $class->GT::Base::new(@_);
    my ($max_bytes, $max_files) = (-1, -1);

    if ($quotastr) {
        croak "Empty quota string specified to $class->open" unless length $quotastr;
        ($max_bytes, $max_files) = _readquotadescr($quotastr);
    }

    $self->{max_bytes} = $max_bytes;
    $self->{max_files} = $max_files;
    $self->{added_bytes} = 0;
    $self->{added_files} = 0;
    $self->{needs_checksize} = 1;
    $self->{has_recalculate} = 0;
    $self->{fh} = do { local *FH; \*FH };

    if (-e "$dir/maildirfolder") {
        $self->{dir} = "$dir/..";
    }
    else {
        $self->{dir} = $dir;
    }

    if (!$quotastr and !_checksize($self, 1) and 
            $self->{max_bytes} < 0 and $self->{max_files} < 0)
    {
        return undef;
    }
    return $self;
}

sub init {
}

sub close {
    my $self = shift;
    if (defined fileno $self->{fh}) {
        if ($self->{added_files} != 0 and $self->{added_bytes} != 0) {
            my $fh = $self->{fh};
            print $fh "$self->{added_bytes} $self->{added_files}\n";
            select((select($fh), $| = 1)[0]);
        }
        CORE::close $self->{fh};
    }
    undef;
}

sub DESTROY {
    my $self = shift;
    $self->close();
}

sub test {
    my $self = shift;
    my $bytes = shift || 0;
    my $files = shift || 0;
    $error = undef;
    if ($self->{needs_checksize}) {
        _checksize($self, 0) or return undef;
    }
    if (!_quota_check($self, $bytes, $files)) {
        my @st;
        if ($self->{has_recalculate} == 0 and defined fileno $self->{fh} and
            @st = stat($self->{fh}) and $st[ST_MTIME] + MAILDIRQUOTA_RECALCULATE_DELAY < time)
        {
            if (!_recalculate($self)) {
                return undef;
            }
            if (!_quota_check($self, $bytes, $files)) {
                return undef;
            }
        }
        else {
            return undef;
        }
    }
    return 1;
}

sub get_current_bytes {
    my $self = shift;
    croak "Unknown arguments to $self->get_current_bytes" if @_;
    return undef unless $self->test(0, 0);
    return $self->{bytes} + $self->{added_bytes};
}

sub get_max_bytes {
    my $self = shift;
    croak "Unknown arguments to $self->get_max_bytes" if @_;
    return undef unless $self->test(0, 0);
    return $self->{max_bytes};
}

sub get_current_files {
    my $self = shift;
    croak "Unknown arguments to $self->get_current_files" if @_;
    return undef unless $self->test(0, 0);
    return $self->{files} + $self->{added_files};
}

sub get_max_files {
    my $self = shift;
    croak "Unknown arguments to $self->get_max_files" if @_;
    return undef unless $self->test(0, 0);
    return $self->{max_files};
}

sub add {
    my $self = shift;
    my $bytes = shift || 0;
    my $files = shift || 0;
    croak "Unknonw arguments to $self->add" if @_;
    if ($self->{needs_checksize}) {
        _checksize($self, 0) or return undef;
    }
    $self->{added_bytes} += $bytes;
    $self->{added_files} += $files;
    return 1;
}

sub _quota_check {
    my ($self, $bytes, $files) = @_;
    if ($self->{needs_checksize}) {
        return 1;
    }
    if (($self->{max_bytes} > 0 and $bytes > 0 and
            $self->{bytes} + $self->{added_bytes} + $bytes > $self->{max_bytes}) ||
            ($self->{max_files} > 0 and $files > 0 and
            $self->{files} + $self->{added_files} + $files > $self->{max_files}))
    {
        return undef;
    }
    else {
        return 1;
    }
}

sub _getrealsize {
    my $self = shift;
    my $dir = $self->{dir};
    my $do_bytes = shift;
    my $do_files = shift;
    my $dirp = do { local *FDIR; \*FDIR };
    my $remove_trash = !$Trash_Quota;
    my $bytes = 0;
    my $files = 0;
    my $stamp = 0;
    if (opendir $dirp, $dir) {
        while (my $ent = readdir $dirp) {
            next if $ent eq '.' or $ent eq '..';
            next unless $ent =~ /^[.](.+)/;
            if ($remove_trash and $1 eq $Trash_Folder) {
                $remove_trash = 0;
                next;
            }
            for (qw/new cur/) {
                my $cdir = "$dir/$ent/$_";
                my @st = stat $cdir;
                next unless @st;
                if ($st[ST_MTIME] > $stamp) {
                    $stamp = $st[ST_MTIME];
                }
                if ($do_bytes or $do_files) {
                    my $dir = do { local *SDIR; \*SDIR };
                    opendir $dir, $cdir or return $self->error("OPENDIR", $cdir, "$!");
                    while (my $fent = readdir $dir) {
                        if ($fent =~ /S=(\d+):2,(\w*)/ and index($2, "T") != -1) {
                            $bytes += $1;
                            $files++;
                        }
                        else {
                            my @st = lstat("$cdir/$fent");
                            if (@st and S_ISREG($st[ST_MODE])) {
                                $bytes += $st[7];
                                $files++;
                            }
                        }
                    }
                    closedir $dir;
                }
            }
        }
    }
    closedir $dirp;
    return($stamp, $bytes, $files);
}

sub _readquotadescr {
    my ($quotastr) = @_;
    my ($max_bytes, $max_files) = (-1, -1);
    my @desc = split /,/, $quotastr;
    for (@desc) {
        if (/(\d+)S/) {
            $max_bytes = $1;
        }
        elsif (/(\d+)C/) {
            $max_files = $1;
        }
        else {
            croak "Invalid quota description string: $_";
        }
    }
    return($max_bytes, $max_files);
}

sub _recalculate {
    my ($self) = @_;
    my $saved_time;

    $self->{bytes} = 0;
    $self->{files} = 0;

    # avoid endless recalculations on error
    $self->{needs_checksize} = 0;
    $self->{has_recalculate} = 1;

    if (defined fileno $self->{fh}) {
        CORE::close $self->{fh};
    }

    ($saved_time, $self->{bytes}, $self->{files}) = _getrealsize($self, 1, 1);
    
    my $tmpfile = new GT::TempFile(
        destroy => 0,
        tmp_dir => "$self->{dir}/tmp"
    );
    my $fh = $self->{fh};
    CORE::open $fh, ">$$tmpfile" or return $self->error("WARN", "READOPEN", $$tmpfile, "$!");
    my $buff = '';
    if ($self->{max_bytes} > 0) {
        $buff = "$self->{max_bytes}S";
        if ($self->{max_files} > 0) {
            $buff .= ",";
        }
    }
    if ($self->{max_files} > 0) {
        $buff .= "$self->{max_files}C";
    }
    $buff .= "\n";
    if ($self->{bytes} > 0 and $self->{files} > 0) {
        $buff .= "$self->{bytes} $self->{files}\n";
    }
    print $fh $buff;
    select((select($fh), $|=1)[0]);

    my $maildirsize = "$self->{dir}/maildirsize";
    if (!rename($$tmpfile, $maildirsize)) {
        CORE::close $fh;
        unlink $$tmpfile;
        return $self->error("RENAME", "WARN", $$tmpfile, $maildirsize, "$!");
    }
    my ($time) = _getrealsize($self, 0, 0);
    if ($time > $saved_time) {
        CORE::close $fh;
        unlink $maildirsize;
    }
    return 1;
}

sub _checksize {
    my ($self, $getinfo) = @_;

    my $size_file = "$self->{dir}/maildirsize";

    unless (defined fileno $self->{fh}) {
        if (!CORE::open($self->{fh}, "+< $size_file")) {
            if ($! == 2) { # ENOENT
                if ($getinfo) {
                    return $self->error("OPEN", $size_file, "$!");
                }
                return _recalculate($self);
            }
            else {
                return $self->error("OPEN", $size_file, "$!");
            }
        }
    }
    my $fh = $self->{fh};
    local $!;
    my $line = <$fh>;
    unless (defined $line) {
        CORE::close $fh;
        return _recalculate($self);
    }
    chomp $line;
    my ($max_bytes, $max_files) = _readquotadescr($line);
    if ($getinfo) {
        $self->{max_bytes} = $max_bytes;
        $self->{max_files} = $max_files;
    }
    elsif ($max_bytes != $self->{max_bytes} or $max_files != $self->{max_files}) {
        return _recalculate($self);
    }
    my ($bytes_diff, $files_diff) = (0, 0);
    while (<$fh>) {
        my ($bytes, $files) = split ' ';
        $bytes_diff += $bytes;
        $files_diff += $files;
    }
    $self->{bytes} = $bytes_diff;
    $self->{files} = $files_diff;
    $self->{needs_checksize} = 1;
    return 1;
}

1;

